"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.phys import Fluid

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

VERBOSE = False


class PureFluidTestCase(unittest.TestCase):
    """
    Test engineering toolkit temperature, pressure, density API interface
    """
    def setUp(self):
        pass

    def test_air(self):
        fluid = Fluid(species='air', T=298, P=101325)
        self.assertAlmostEqual(fluid.T, 298.)
        self.assertAlmostEqual(fluid.P, 101325.)

        fluid = Fluid(species='air', T=290, P=103000)
        self.assertAlmostEqual(fluid.T, 290.)
        self.assertAlmostEqual(fluid.P, 103000)

    def test_bad_air(self):
        bad_name = 'ari'
        with self.assertRaises(ValueError):
            _ = Fluid(species=bad_name, T=290, P=103000)

    def test_hydrogen(self):
        fluid = Fluid(species='hydrogen', T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species='h2', T=315, P=200000)
        self.assertAlmostEqual(fluid.T, 315.)
        self.assertAlmostEqual(fluid.P, 200000.)

    def test_hydrogen_dict(self):
        fluid = Fluid(species={'hydrogen': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species={'h2': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

    def test_methane(self):
        fluid = Fluid(species='methane', T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species='ch4', T=315, P=200000)
        self.assertAlmostEqual(fluid.T, 315.)
        self.assertAlmostEqual(fluid.P, 200000.)

    def test_methane_dict(self):
        fluid = Fluid(species={'methane': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species={'ch4': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

    def test_propane(self):
        fluid = Fluid(species='propane', T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species='c3h8', T=315, P=200000)
        self.assertAlmostEqual(fluid.T, 315.)
        self.assertAlmostEqual(fluid.P, 200000.)

    def test_propane_dict(self):
        fluid = Fluid(species={'propane': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

        fluid = Fluid(species={'c3h8': 1.0}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)


class BlendFluidTestCase(unittest.TestCase):
    """
    Test engineering toolkit temperature, pressure, density API interface.

    """
    def setUp(self):
        pass

    def test_methane_blend(self):
        fluid = Fluid(species={'ch4': 0.965, 'n2': 0.035}, T=300, P=110000)
        self.assertAlmostEqual(fluid.T, 300.)
        self.assertAlmostEqual(fluid.P, 110000.)

    def test_insufficient_blend_raises_error(self):
        with self.assertRaises(ValueError):
            _ = Fluid(species={'methane': 0.5, 'h2': 0.4}, T=300, P=110000)

    def test_h2_methane_blend_with_default_t_p(self):
        fluid = Fluid(species={'ch4': 0.5, 'n2': 0.5}, T=287, P=35e6)
        self.assertAlmostEqual(fluid.T, 287.)
        self.assertAlmostEqual(fluid.P, 35e6)




"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
import json

import numpy as np
from scipy.constants import milli, kilo, liter, bar, pi
from scipy import optimize
from scipy import interpolate

from hyram.phys import Source, Fluid, Orifice, Flame
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-heatflux')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-heatflux.json')


class Test_Ekoto_2014(unittest.TestCase):
    """
    Test suite for jet flame length and heat flux, data from Ekoto, Ruggles, Creitz, Li (2014)
    """

    def setUp(self):
        ambient_pressures = np.array([1.022, 1.011])*bar
        ambient_temperatures = np.array([280, 280])#K
        diameters = np.array([20.9, 52.5])*milli#m
        pressures = np.array([59.8, 62.1])*bar + ambient_pressures
        temperatures = np.array([308.7, 287.8]) #K
        self.humidities = [0.943, 0.945]
        y0 = 3.25
        self.rad_x = [26, 48]
        self.rad_y = 1.75
        self.rad_z = 0

        self.flames = []
        for amb_pressure, amb_temp, diameter, pressure, temp in zip(ambient_pressures, ambient_temperatures, diameters, pressures, temperatures):
            air = Fluid(species='air', T=amb_temp, P=amb_pressure)
            gas = Fluid(species='H2', T=temp, P=pressure)
            orifice = Orifice(d=diameter)

            self.flames.append(Flame(gas, orifice, air, y0=y0))

        # Get error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Flame length
    def test_flame_length(self):
        title = "Ekoto et al. 2014 - Table 2, Flame Length"
        exp_lengths = [17.4, 45.9]
        calc_lengths = [flame.length() for flame in self.flames]
        error_limits = self.limits[title]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_lengths,
                                calc_vals=calc_lengths,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")


    # Heat flux
    def test_flux(self):
        title = "Ekoto et al. 2014 - Table 2, Heat Flux"
        exp_flux = [4.7, 23.9]
        calc_flux = [flame.Qrad_multi(x, self.rad_y, self.rad_z, RH=RH) for flame, x, RH in zip(self.flames, self.rad_x, self.humidities)]
        calc_flux_adj = [1e-3 * x for x in calc_flux] # convert to kW/m^2
        error_limits = self.limits[title]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux_adj,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")


class Test_Schefer_2006_2007(unittest.TestCase):
    """
    Test suite for Heat flux vs. Distance and Flame length vs. Time.
    Data from  Schefer, Houf, Bourne, Colton (2006) and Houf and Schefer (2007)
    """

    def setUp(self):
        diameter = 7.94*milli#m
        temp = 270 #K - assumed
        amb_pressure = 101325 # assumed
        amb_temp = 293 #K - assumed
        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        orifice = Orifice(d=diameter)

        # Keys are time t=5s, 20, 40, 60, 70
        init_flowrates = {5:57.3, 20:23.17, 40:6.92, 60:2.07, 70:1.13} #g/s
        velocities = {5:1233, 20:1231, 40:1078, 60:644, 70:446} #m/s
        densities = dict([[t, init_flowrates[t]/kilo/(velocities[t]*pi/4*diameter**2)] for t in init_flowrates.keys()]) #g/m^3
        gases, flowrates = {}, {}
        for t in densities.keys():
            gases[t] = Fluid(species='H2', T=temp, rho=densities[t])

            if gases[t].P <= 2 * amb_pressure: # unchoked
                gases[t] = Fluid(species='H2', T=temp, P=amb_pressure)
                flowrates[t] = init_flowrates[t] / kilo
            else: # choked - solve for stagnation density
                def err_mdot(rho0):
                    gases[t] = Fluid(species='H2', T=temp, rho=rho0[0])
                    return orifice.mdot(orifice.flow(gases[t], amb_pressure)) - init_flowrates[t] / kilo
                rho0 = optimize.root(err_mdot, densities[t])['x'][0]
                gases[t] = Fluid(species='H2', T=temp, rho=rho0)
                flowrates[t] = None

        self.flames = []
        for t in gases.keys():
            self.flames.append(Flame(gases[t], orifice, air, mdot=flowrates[t], theta0=pi/2))

        # Get error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Heat Flux / Normalized Distance
    def test_flux_distance_fig8(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'schefer-2006-fig8.csv'), header=2)

        times = [5, 20, 40, 60, 70]
        for i in range(len(times)):
            title = f"Schefer 2006 - Figure 8, Flux v. Normalized Distance at t={times[i]}"
            error_limits = self.limits[title]

            # Data is under x1/y1, x3/y3, etc
            idx = (2 * i) + 1
            distance = data[f'x{idx}']
            exp_flux = [d * 1e3 for d in data[f'y{idx}']] # Convert kW to W

            flame_length = self.flames[i].length()
            y = [d * flame_length for d in distance]
            x = 1.82 * np.ones_like(y)
            z = np.zeros_like(y)
            calc_flux = self.flames[i].Qrad_multi(x, y, z, RH = .9)

            # Calculate error statistics and test
            error = utils.get_error(exp_vals=exp_flux,
                                    calc_vals=calc_flux,
                                    error_limits=error_limits,
                                    units='W/m^2',
                                    msg=title,
                                    output_filename=OUTPUT_FILE,
                                    create_output=CREATE_OUTPUT,
                                    verbose=VERBOSE)

            with self.subTest():
                self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
            with self.subTest():
                self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
            with self.subTest():
                self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

            # Create plots
            if CREATE_PLOTS:
                utils.create_plots(output_dir=OUTPUT_LOC,
                                   x=distance,
                                   exp_y=exp_flux,
                                   calc_y=calc_flux,
                                   max_error=error['Max Absolute Error'],
                                   avg_error=error['Avg Absolute Error'],
                                   smape=error['SMAPE'],
                                   title=title,
                                   xlabel=r'Normalized Distance ($x/L_{\rm vis}$)',
                                   ylabel='Heat Flux (W/m$^2$)')

    # Test figure 6
    def test_flux_distance_fig6(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'houfschefer-2007-fig6.csv'), header=2)
        title = "Houf & Shefer 2007 - Figure 6, Flux v. Normalized Distance"
        error_limits = self.limits[title]

        distance = data['x1']
        exp_flux = [d * 1e3 for d in data['y1']] # Convert kW to W

        flame_length = self.flames[0].length()
        y = [d * flame_length for d in distance]
        x = 1.82*np.ones_like(y)
        z = np.zeros_like(y)
        calc_flux = self.flames[0].Qrad_multi(x, y, z, .9)

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")

        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=distance,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel=r'Normalized Distance ($x/L_{\rm vis}$)',
                               ylabel='Heat Flux (W/m$^2$)')

    # Flame length / time
    def test_length_time(self):
        data = utils.read_csv(os.path.join(DATA_LOC, 'houfschefer-2007-fig5-ADJUSTED.csv'), header=2)
        title = "Houf & Shefer 2007 - Figure 5, Flame Length v. Time"
        error_limits = self.limits[title]

        exp_time = data['x1']
        exp_length = data['y1']

        calc_time = [5, 20, 40, 60, 70]
        calc_length = [flame.length() for flame in self.flames]

        # Interpolate calculated pressure
        interp = interpolate.interp1d(exp_time, exp_length, fill_value='extrapolate')
        exp_length = interp(calc_time)

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Flame Length (m)')


class Test_Imamura_2008(unittest.TestCase):
    """
    Test suite for Mass Flow Rate and Flame Rate vs. Pressure.
    Data from Imamura et al. (2008)
    """

    def setUp(self):
        amb_pressure = 101325 # assumed
        amb_temp = 293 #K - assumed
        air = Fluid(T = amb_temp, P = amb_pressure, species = 'air')

        self.data = dict({'d_mm':[1,1,1,2,2,2,2,2,2,3,3,3,3,3,3,4,4,4,4,4,4],
                          'P_MPa':[1.09,2.17,3.26,0.47,0.93,1.4,1.87,2.34,2.8,0.36,0.73,1.09,1.46,1.82,2.19,0.23,0.47,0.7,0.94,1.17,1.41],
                          'Flow':[0.000536094, 0.001031305, 0.001567413, 0.000902573, 0.001805217, 0.00272147, 0.003678693,
                                0.004581309, 0.005470301, 0.001516577, 0.003142413, 0.004713612, 0.006298463, 0.007910646,
                                0.009468193, 0.00181646, 0.003673918, 0.005449494, 0.007334199, 0.00910976, 0.010967247],
                          'Lf_m':[0.35,0.6,0.7,0.55,0.72,0.99,1.04,1.29,1.29,0.78,1.08,1.31,1.45,1.62,1.67,0.95,1.21,1.49,1.61,1.77,1.82]})

        flames =[]
        for d in range(len(self.data['d_mm'])):
            orifice = Orifice(d=self.data['d_mm'][d]*milli)
            pressure = amb_pressure + self.data['P_MPa'][d] * 1e6
            gas = Fluid(species='H2', T=amb_temp, P=pressure)
            flames.append(Flame(gas, orifice, air))

        self.data['Flames'] = flames

        # Create subsets of the data for each d value
        self.data_1mm = dict()
        self.data_2mm = dict()
        self.data_3mm = dict()
        self.data_4mm = dict()

        for key in self.data:
            self.data_1mm[key] = self.data[key][:3]
            self.data_2mm[key] = self.data[key][3:9]
            self.data_3mm[key] = self.data[key][9:15]
            self.data_4mm[key] = self.data[key][15:21]

        # Get error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Test Mass flowrate vs. Pressure, d = 1mm
    def test_flowrate_1mm(self):
        title = "Imamura et al. 2008 - Figure 3, Mass Flow Rate v. Pressure, d=1mm"
        error_limits = self.limits[title]

        pressure = self.data_1mm['P_MPa']
        exp_flowrate = self.data_1mm['Flow']
        calc_flowrate = [f.mass_flow_rate for f in self.data_1mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flowrate,
                                calc_vals=calc_flowrate,
                                error_limits=error_limits,
                                units='kg/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_flowrate,
                               calc_y=calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Mass Flow Rate (kg/s)')

    # Test Mass flowrate vs. Pressure, d = 2mm
    def test_flowrate_2mm(self):
        title = "Imamura et al. 2008 - Figure 3, Mass Flow Rate v. Pressure, d=2mm"
        error_limits = self.limits[title]

        pressure = self.data_2mm['P_MPa']
        exp_flowrate = self.data_2mm['Flow']
        calc_flowrate = [f.mass_flow_rate for f in self.data_2mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flowrate,
                                calc_vals=calc_flowrate,
                                error_limits=error_limits,
                                units='kg/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_flowrate,
                               calc_y=calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Mass Flow Rate (kg/s)')

    # Test Mass flowrate vs. Pressure, d = 3mm
    def test_flowrate_3mm(self):
        title = "Imamura et al. 2008 - Figure 3, Mass Flow Rate v. Pressure, d=3mm"
        error_limits = self.limits[title]

        pressure = self.data_3mm['P_MPa']
        exp_flowrate = self.data_3mm['Flow']
        calc_flowrate = [f.mass_flow_rate for f in self.data_3mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flowrate,
                                calc_vals=calc_flowrate,
                                error_limits=error_limits,
                                units='kg/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")

        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_flowrate,
                               calc_y=calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Mass Flow Rate (kg/s)')

    # Test Mass flowrate vs. Pressure, d = 4mm
    def test_flowrate_4mm(self):
        title = "Imamura et al. 2008 - Figure 3, Mass Flow Rate v. Pressure, d=4mm"
        error_limits = self.limits[title]

        pressure = self.data_4mm['P_MPa']
        exp_flowrate = self.data_4mm['Flow']
        calc_flowrate = [f.mass_flow_rate for f in self.data_4mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_flowrate,
                                calc_vals=calc_flowrate,
                                error_limits=error_limits,
                                units='kg/s',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_flowrate,
                               calc_y=calc_flowrate,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Mass Flow Rate (kg/s)')


    # Test Flame Length vs. Pressure, d = 1mm
    def test_flamelength_1mm(self):
        title = "Imamura et al. 2008 - Figure 3, Flame Length v. Pressure, d=1mm"
        error_limits = self.limits[title]

        pressure = self.data_1mm['P_MPa']
        exp_length = self.data_1mm['Lf_m']
        calc_length = [f.length() for f in self.data_1mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Flame Length (m)')

    # Test Flame Length vs. Pressure, d = 2mm
    def test_flamelength_2mm(self):
        title = "Imamura et al. 2008 - Figure 3, Flame Length v. Pressure, d=2mm"
        error_limits = self.limits[title]

        pressure = self.data_2mm['P_MPa']
        exp_length = self.data_2mm['Lf_m']
        calc_length = [f.length() for f in self.data_2mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Flame Length (m)')

    # Test Flame Length vs. Pressure, d = 3mm
    def test_flamelength_3mm(self):
        title = "Imamura et al. 2008 - Figure 3, Flame Length v. Pressure, d=3mm"
        error_limits = self.limits[title]

        pressure = self.data_3mm['P_MPa']
        exp_length = self.data_3mm['Lf_m']
        calc_length = [f.length() for f in self.data_3mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Flame Length (m)')

    # Test Flame Length vs. Pressure, d = 4mm
    def test_flamelength_4mm(self):
        title = "Imamura et al. 2008 - Figure 3, Flame Length v. Pressure, d=4mm"
        error_limits = self.limits[title]

        pressure = self.data_4mm['P_MPa']
        exp_length = self.data_4mm['Lf_m']
        calc_length = [f.length() for f in self.data_4mm['Flames']]

        # Calculate error statistics and test
        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (MPa)',
                               ylabel='Flame Length (m)')


class Test_Mogi_2005(unittest.TestCase):
    """
    Test suite for Heat Flux vs. Mass Flow Rate.
    Data from Mogi, Nishida, Horiguci (2005)
    """

    def setUp(self):
        amb_pressure = 101325 # assumed
        amb_temp = 293 #K - assumed
        air = Fluid(T=amb_temp, P=amb_pressure, species='air')
        self.mogi_flames = {'d':np.array([.4, .8, 2])*milli,
                            'P':np.array([35, 25, 15, 5, 1])*1e6,
                            'L':[1.5, 2.5, 3.5]}

        # Calculate flames and add to mogi_flames
        flames = []
        for diameter in self.mogi_flames['d']:
            for pressure in self.mogi_flames['P']:
                orifice = Orifice(diameter)
                gas = Fluid(species='H2', T=amb_temp, P=(amb_pressure + pressure))
                flames.append(Flame(gas, orifice, air))

        self.mogi_flames['Flames'] = flames

        self.data = utils.read_csv(os.path.join(DATA_LOC, 'mogi-2005-fig9.csv'), header=3)

        # Get error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Mass Flow Rate vs. Heat Flux, L = 1.5m (distance from detector to flame axis)
    def test_flowrate_flux_L1(self):
        title = "Mogi et al. 2005 - Figure 9, Heat Flux v. Mass Flow Rate, L=1.5m"
        error_limits = self.limits[title]

        L = 1.5
        calc_flowrate = []
        calc_flux = []
        for i in range(len(self.mogi_flames['Flames'])):
            flame = self.mogi_flames['Flames'][i]

            flame_length = np.linspace(0, flame.length())
            calc_flowrate.append(flame.developing_flow.orifice.mdot(flame.developing_flow.fluid_orifice))
            calc_flux.append(max(flame.Qrad_multi(flame_length, np.ones_like(flame_length)*1, np.ones_like(flame_length)*L, .9)))

        # Sort by flowrate (X)
        calc_flux = [flux for _, flux in sorted(zip(calc_flowrate, calc_flux))]
        calc_flowrate = sorted(calc_flowrate)

        # Get experimental data
        exp_flowrate = self.data['x1']
        exp_flowrate.extend(self.data['x2'])
        exp_flowrate.extend(self.data['x3'])

        exp_flux = self.data['y1']
        exp_flux.extend(self.data['y2'])
        exp_flux.extend(self.data['y3'])

        # Sort data
        exp_flux = [flux for _, flux in sorted(zip(exp_flowrate, exp_flux))]
        exp_flowrate = sorted(exp_flowrate)

        # Interpolate calculated pressure
        interp = interpolate.interp1d(exp_flowrate, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_flowrate)

        # Get error values and test
        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_flowrate,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Mass Flow Rate (kg/s)',
                               ylabel='Heat Flux (W/m$^2$)')

    # Mass Flow Rate vs. Heat Flux, L = 2.5m (distance from detector to flame axis)
    def test_flowrate_flux_L2(self):
        title = "Mogi et al. 2005 - Figure 9, Heat Flux v. Mass Flow Rate, L=2.5m"
        error_limits = self.limits[title]

        L = 2.5
        calc_flowrate = []
        calc_flux = []
        for i in range(len(self.mogi_flames['Flames'])):
            flame = self.mogi_flames['Flames'][i]

            l = np.linspace(0, flame.length())
            calc_flowrate.append(flame.developing_flow.orifice.mdot(flame.developing_flow.fluid_orifice))
            calc_flux.append(max(flame.Qrad_multi(l, np.ones_like(l)*1, np.ones_like(l)*L, .9)))

        # Sort by flowrate (X)
        calc_flux = [flux for _, flux in sorted(zip(calc_flowrate, calc_flux))]
        calc_flowrate = sorted(calc_flowrate)

        # Get experimental data
        exp_flowrate = self.data['x5']
        exp_flowrate.extend(self.data['x6'])
        exp_flowrate.extend(self.data['x7'])

        exp_flux = self.data['y5']
        exp_flux.extend(self.data['y6'])
        exp_flux.extend(self.data['y7'])

        # Sort data
        exp_flux = [flux for _, flux in sorted(zip(exp_flowrate, exp_flux))]
        exp_flowrate = sorted(exp_flowrate)

        interp = interpolate.interp1d(exp_flowrate, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_flowrate)

        # Get error values and test
        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_flowrate,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Mass Flow Rate (kg/s)',
                               ylabel='Heat Flux (W/m$^2$)')


    # Mass Flow Rate vs. Heat Flux, L = 3.5m (distance from detector to flame axis)
    def test_flowrate_flux_L3(self):
        title = "Mogi et al. 2005 - Figure 9, Heat Flux v. Mass Flow Rate, L=3.5m"
        error_limits = self.limits[title]

        L = 3.5
        calc_flowrate = []
        calc_flux = []
        for i in range(len(self.mogi_flames['Flames'])):
            flame = self.mogi_flames['Flames'][i]

            l = np.linspace(0, flame.length())
            calc_flowrate.append(flame.developing_flow.orifice.mdot(flame.developing_flow.fluid_orifice))
            calc_flux.append(max(flame.Qrad_multi(l, np.ones_like(l)*1, np.ones_like(l)*L, .9)))

        # Sort by flowrate (X)
        calc_flux = [flux for _, flux in sorted(zip(calc_flowrate, calc_flux))]
        calc_flowrate = sorted(calc_flowrate)

        # Get experimental data
        exp_flowrate = self.data['x9']
        exp_flowrate.extend(self.data['x10'])
        exp_flowrate.extend(self.data['x11'])

        exp_flux = self.data['y9']
        exp_flux.extend(self.data['y10'])
        exp_flux.extend(self.data['y11'])

        # Sort data
        exp_flux = [flux for _, flux in sorted(zip(exp_flowrate, exp_flux))]
        exp_flowrate = sorted(exp_flowrate)

        interp = interpolate.interp1d(exp_flowrate, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_flowrate)

        # Get error values and test
        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_flowrate,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Mass Flow Rate (kg/s)',
                               ylabel='Heat Flux (W/m$^2$)')


class Test_Proust_2011(unittest.TestCase):
    """
    Variety of tests, including Pressure and Tank measurements.
    Data from Proust, Jamois, Studer (2011)
    """

    def setUp(self):
        self.y0 = 1.5 #m
        amb_pressure = 101325 # assumed
        amb_temp = 293 #K - assumed
        tank_pressure = 90*1e6 #Pa
        tank_volume = 25*liter
        diameters = np.array([1,2,3])*milli #m

        self.pressures = np.array([900,800,700,600,500,400,300,250,200,175,150,125,100,80,60,40,30,20]) #bar
        self.pressures_2 = np.append(self.pressures, [15, 10, 5, 4, 3, 2])
        self.temp_1mm = np.array([42,40,36,27,15,-2,-18,-25,-31,-34,-37,-38,-39,-39,-37,-34,-32,-28]) #C
        self.density_1mm = np.array([42,40,37,34,31,27,23,20,17,15,13,11,9,8,6,4,3,2])
        self.temp_2mm = np.array([42,40,38,34,27,15,-4,-20,-39,-47,-54,-58,-62,-64,-64,-62,-60,-56])
        self.density_2mm = np.array([41,39,37,34,31,27,23,20,17,15,13,11,9,7,5,3,3,2])
        self.temp_3mm = np.array([46,44,42,39,35,28,15,4,-11,-20,-30,-47,-61,-70,-77,-80,-78,-76])
        self.density_3mm = np.array([42,39,37,34,31,28,23,21,17,15,14,11,9,7,5,3,3,2])

        gas = Fluid(species='H2', T = self.temp_2mm[0]+273, P=(amb_pressure+tank_pressure))
        source = Source(tank_volume, gas)
        self.air = Fluid(T = amb_temp, P = amb_pressure, species = 'air')

        self.blowdown_data = []
        for d in diameters:
            orifice = Orifice(d)
            self.blowdown_data.append(source.empty(orifice))

        # Importing figure data
        self.fig4data = utils.read_csv(os.path.join(DATA_LOC, 'proust-2011-fig4.csv'), header=1)
        self.fig8data = utils.read_csv(os.path.join(DATA_LOC, 'proust-2011-fig8.csv'), header=2)
        self.fig10data = utils.read_csv(os.path.join(DATA_LOC, 'proust-2011-fig10.csv'), header=2)

        # Get error limits
        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

    # Pressure vs. Time, fig. 4
    def test_pressure_time(self):
        title = "Proust et al. 2011 - Pressure v. Time, d=2mm"
        error_limits = self.limits[title]

        # Comparing 2mm data
        exp_time = self.fig4data['x3']
        exp_pressure = self.fig4data['y3']

        calc_time = self.blowdown_data[1][2]
        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[1][1]]
        interp = interpolate.interp1d(exp_time, exp_pressure, fill_value='extrapolate')
        exp_pressure = interp(calc_time)

        # Get error values and test
        error = utils.get_error(exp_vals=exp_pressure,
                                calc_vals=calc_pressure,
                                error_limits=error_limits,
                                units='bar',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_pressure,
                               calc_y=calc_pressure,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Pressure (bar)')


    # Tank Pressure vs. Temperature, 1mm, fig. 4a
    def test_temp_pressure_1mm(self):
        exp_pressure = self.pressures
        exp_temp = self.temp_1mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[0][1]]
        calc_temp = [g.T-273 for g in self.blowdown_data[0][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_temp), fill_value='extrapolate')
        exp_temp = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 4a, Temperature v. Pressure, d=1mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_temp,
                                calc_vals=calc_temp,
                                error_limits=error_limits,
                                units='C',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_temp,
                               calc_y=calc_temp,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Tank Temperature (C)')

    # Tank Pressure vs. Tank Temperature, 2mm, fig. 4a
    def test_temp_pressure_2mm(self):
        exp_pressure = self.pressures
        exp_temp = self.temp_2mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[1][1]]
        calc_temp = [g.T-273 for g in self.blowdown_data[1][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_temp), fill_value='extrapolate')
        exp_temp = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 4a, Temperature v. Pressure, d=2mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_temp,
                                calc_vals=calc_temp,
                                error_limits=error_limits,
                                units='C',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_temp,
                               calc_y=calc_temp,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Tank Temperature (C)')

    # Tank Pressure vs. Tank Temperature, 3mm, fig. 4a
    def test_temp_pressure_3mm(self):
        exp_pressure = self.pressures
        exp_temp = self.temp_3mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[2][1]]
        calc_temp = [g.T-273 for g in self.blowdown_data[2][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_temp), fill_value='extrapolate')
        exp_temp = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 4a, Temperature v. Pressure, d=3mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_temp,
                                calc_vals=calc_temp,
                                error_limits=error_limits,
                                units='C',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_temp,
                               calc_y=calc_temp,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Tank Temperature (C)')


    # Tank Pressure vs. Density, 1mm, fig. 5
    def test_density_pressure_1mm(self):
        exp_pressure = self.pressures
        exp_density = self.density_1mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[0][1]]
        calc_density = [g.rho for g in self.blowdown_data[0][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_density), fill_value='extrapolate')
        exp_density = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 5, Density v. Pressure, d=1mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_density,
                                calc_vals=calc_density,
                                error_limits=error_limits,
                                units='kg/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_density,
                               calc_y=calc_density,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Density (kg/$m^2$)')

    # Tank Pressure vs. Density, 2mm, fig. 5
    def test_density_pressure_2mm(self):
        exp_pressure = self.pressures
        exp_density = self.density_1mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[1][1]]
        calc_density = [g.rho for g in self.blowdown_data[1][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_density), fill_value='extrapolate')
        exp_density = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 5, Density v. Pressure, d=2mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_density,
                                calc_vals=calc_density,
                                error_limits=error_limits,
                                units='kg/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_density,
                               calc_y=calc_density,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Density (kg/$m^2$)')


    # Tank Pressure vs. Density, 3mm, fig. 5
    def test_density_pressure_3mm(self):
        exp_pressure = self.pressures
        exp_density = self.density_1mm

        calc_pressure = [g.P*1e-5 for g in self.blowdown_data[2][1]]
        calc_density = [g.rho for g in self.blowdown_data[2][1]]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_density), fill_value='extrapolate')
        exp_density = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 5, Density v. Pressure, d=3mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_density,
                                calc_vals=calc_density,
                                error_limits=error_limits,
                                units='kg/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_density,
                               calc_y=calc_density,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Tank Pressure (bar)',
                               ylabel='Density (kg/$m^2$)')


    # Pressure vs. Flame Length, 1mm, fig. 8
    def test_length_pressure_1mm(self):
        # Build flames
        temps = self.temp_1mm + 273
        orifice = Orifice(1*milli)

        flames = []
        for i, p in enumerate(self.pressures):
            gas = Fluid(species='H2', T = temps[i], P = p*1e5)
            flames.append(Flame(gas, orifice, self.air, y0=self.y0))

        exp_pressure = self.fig8data['x3']
        exp_length = self.fig8data['y3']

        calc_pressure = self.pressures
        calc_length = [f.length() for f in flames]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_length), fill_value='extrapolate')
        exp_length = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 8, Flame Length v. Pressure, d=1mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (bar)',
                               ylabel='Flame Length (m)')

    # Pressure vs. Flame Length, 2mm, fig. 8
    def test_length_pressure_2mm(self):
        # Build flames
        temps = self.temp_2mm + 273
        orifice = Orifice(2*milli)

        flames = []
        for i, p in enumerate(self.pressures):
            gas = Fluid(species='H2', T = temps[i], P = p*1e5)
            flames.append(Flame(gas, orifice, self.air, y0=self.y0))

        exp_pressure = self.fig8data['x2']
        exp_length = self.fig8data['y2']

        calc_pressure = self.pressures
        calc_length = [f.length() for f in flames]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_length), fill_value='extrapolate')
        exp_length = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 8, Flame Length v. Pressure, d=2mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (bar)',
                               ylabel='Flame Length (m)')

    # Pressure vs. Flame Length, 3mm, fig. 8
    def test_length_pressure_3mm(self):
        # Build flames
        temps = self.temp_3mm + 273
        orifice = Orifice(3*milli)

        flames = []
        for i, p in enumerate(self.pressures):
            gas = Fluid(species='H2', T = temps[i], P = p*1e5)
            flames.append(Flame(gas, orifice, self.air, y0=self.y0))

        exp_pressure = self.fig8data['x1']
        exp_length = self.fig8data['y1']

        calc_pressure = self.pressures
        calc_length = [f.length() for f in flames]

        interp = interpolate.interp1d(np.flip(exp_pressure), np.flip(exp_length), fill_value='extrapolate')
        exp_length = interp(calc_pressure)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 8, Flame Length v. Pressure, d=3mm"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_length,
                                calc_vals=calc_length,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_pressure,
                               exp_y=exp_length,
                               calc_y=calc_length,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Pressure (bar)',
                               ylabel='Flame Length (m)')


    # Heat Flux over time #1, fig. 10
    def test_flux_1(self):
        # Get experimental values
        exp_time = self.fig10data['x1']
        exp_flux = self.fig10data['y1']

        # Get Hyram calculated values
        calc_time = [np.interp(P*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], self.blowdown_data[1][2][::-1]) for P in self.pressures_2]
        orifice = Orifice(2*milli) #d = 2mm
        rel_humidity = 0.9

        # Generate flames and calculate heat flux
        rad_loc = 1
        x = rad_loc * 1.2
        y = self.y0
        z = rad_loc
        calc_flux = []
        for i, p in enumerate(self.pressures_2):
            gas = Fluid(species='H2', T = np.interp(self.pressures_2[i]*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], [g.T for g in self.blowdown_data[1][1]][::-1]), P=p*1e5)
            flame = Flame(gas, orifice, self.air, y0=self.y0)
            calc_flux.append(flame.Qrad_multi(x, y, z, rel_humidity))

        interp = interpolate.interp1d(exp_time, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_time)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 10, Heat Flux v. Time, Fluxmeter 1"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Heat Flux (W/$m^2$)')

    # Heat Flux over time #2, fig. 10
    def test_flux_2(self):
        # Get experimental values
        exp_time = self.fig10data['x2']
        exp_flux = self.fig10data['y2']

        # Get Hyram calculated values
        calc_time = [np.interp(P*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], self.blowdown_data[1][2][::-1]) for P in self.pressures_2]
        orifice = Orifice(2*milli) #d = 2mm
        rel_humidity = 0.9

        # Generate flames and calculate heat flux
        rad_loc = 1.5
        x = rad_loc * 1.2
        y = self.y0
        z = rad_loc
        calc_flux = []
        for i, p in enumerate(self.pressures_2):
            gas = Fluid(species='H2', T = np.interp(self.pressures_2[i]*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], [g.T for g in self.blowdown_data[1][1]][::-1]), P=p*1e5)
            flame = Flame(gas, orifice, self.air, y0=self.y0)
            calc_flux.append(flame.Qrad_multi(x, y, z, rel_humidity))

        interp = interpolate.interp1d(exp_time, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_time)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 10, Heat Flux v. Time, Fluxmeter 2"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Heat Flux (W/$m^2$)')

    # Heat Flux over time #3, fig. 10
    def test_flux_3(self):
        # Get experimental values
        exp_time = self.fig10data['x3']
        exp_flux = self.fig10data['y3']

        # Get Hyram calculated values
        calc_time = [np.interp(P*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], self.blowdown_data[1][2][::-1]) for P in self.pressures_2]
        orifice = Orifice(2*milli) #d = 2mm
        rel_humidity = 0.9

        # Generate flames and calculate heat flux
        rad_loc = 2
        x = rad_loc * 1.2
        y = self.y0
        z = rad_loc
        calc_flux = []
        for i, p in enumerate(self.pressures_2):
            gas = Fluid(species='H2', T = np.interp(self.pressures_2[i]*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], [g.T for g in self.blowdown_data[1][1]][::-1]), P=p*1e5)
            flame = Flame(gas, orifice, self.air, y0=self.y0)
            calc_flux.append(flame.Qrad_multi(x, y, z, rel_humidity))

        interp = interpolate.interp1d(exp_time, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_time)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 10, Heat Flux v. Time, Fluxmeter 3"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Heat Flux (W/$m^2$)')


    # Heat Flux over time #4, fig. 10
    def test_flux_4(self):
        # Get experimental values
        exp_time = self.fig10data['x4']
        exp_flux = self.fig10data['y4']

        # Get Hyram calculated values
        calc_time = [np.interp(P*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], self.blowdown_data[1][2][::-1]) for P in self.pressures_2]
        orifice = Orifice(2*milli) #d = 2mm
        rel_humidity = 0.9

        # Generate flames and calculate heat flux
        rad_loc = 3
        x = rad_loc * 1.2
        y = self.y0
        z = rad_loc
        calc_flux = []
        for i, p in enumerate(self.pressures_2):
            gas = Fluid(species='H2', T = np.interp(self.pressures_2[i]*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], [g.T for g in self.blowdown_data[1][1]][::-1]), P=p*1e5)
            flame = Flame(gas, orifice, self.air, y0=self.y0)
            calc_flux.append(flame.Qrad_multi(x, y, z, rel_humidity))

        interp = interpolate.interp1d(exp_time, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_time)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 10, Heat Flux v. Time, Fluxmeter 4"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Heat Flux (W/$m^2$)')


    # Heat Flux over time #5, fig. 10
    def test_flux_5(self):
        # Get experimental values
        exp_time = self.fig10data['x5']
        exp_flux = self.fig10data['y5']

        # Get Hyram calculated values
        calc_time = [np.interp(P*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], self.blowdown_data[1][2][::-1]) for P in self.pressures_2]
        orifice = Orifice(2*milli) #d = 2mm
        rel_humidity = 0.9

        # Generate flames and calculate heat flux
        rad_loc = 4
        x = rad_loc * 1.2
        y = self.y0
        z = rad_loc
        calc_flux = []
        for i, p in enumerate(self.pressures_2):
            gas = Fluid(species='H2', T = np.interp(self.pressures_2[i]*1e5, [g.P for g in self.blowdown_data[1][1]][::-1], [g.T for g in self.blowdown_data[1][1]][::-1]), P=p*1e5)
            flame = Flame(gas, orifice, self.air, y0=self.y0)
            calc_flux.append(flame.Qrad_multi(x, y, z, rel_humidity))

        interp = interpolate.interp1d(exp_time, exp_flux, fill_value='extrapolate')
        exp_flux = interp(calc_time)

        # Get error values and test
        title = "Proust et al. 2011 - Figure 10, Heat Flux v. Time, Fluxmeter 5"
        error_limits = self.limits[title]

        error = utils.get_error(exp_vals=exp_flux,
                                calc_vals=calc_flux,
                                error_limits=error_limits,
                                units='W/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

        # Create plots
        if CREATE_PLOTS:
            utils.create_plots(output_dir=OUTPUT_LOC,
                               x=calc_time,
                               exp_y=exp_flux,
                               calc_y=calc_flux,
                               max_error=error['Max Absolute Error'],
                               avg_error=error['Avg Absolute Error'],
                               smape=error['SMAPE'],
                               title=title,
                               xlabel='Time (s)',
                               ylabel='Heat Flux (W/$m^2$)')


if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.

"""
import unittest

import numpy as np

from hyram.phys import TNT_method, Orifice, Jet, Fluid, BST_method, Bauwens_method
import hyram.phys._unconfined_overpressure as hyram_overp
from hyram.utilities import misc_utils

VERBOSE = False


class OverpressureBlendInputTestCase(unittest.TestCase):
    """ Test unconfined overpressure with blend-like fuel inputs. """
    def setUp(self):
        self.t_amb = 300  # K
        orifice_diameter = 0.0254  # m
        dischage_coefficent = 1
        ambient_pressure = 101325  # Pa

        self.orifice = Orifice(orifice_diameter, dischage_coefficent)
        self.fluid_amb = Fluid(species='air', P=ambient_pressure, T=self.t_amb)

    def test_no_blend_origin_at_leak_orifice(self):
        p_rel = 60e5  # Pa
        release_fluid = Fluid(species='hydrogen', T=self.t_amb, P=p_rel)
        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', release_fluid)
        jet_object = Jet(release_fluid, self.orifice, self.fluid_amb,
                         nn_conserve_momentum=nozzle_cons_m,
                         nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        generic_calc = hyram_overp.Generic_overpressure_method(jet_object, origin_at_orifice=True)
        for coord in generic_calc.origin:
            self.assertAlmostEqual(coord, 0)

    def test_h2(self):
        p_rel = 60e5  # Pa
        release_fluid = Fluid(species={'h2': 1}, T=self.t_amb, P=p_rel)
        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', release_fluid)
        jet_object = Jet(release_fluid, self.orifice, self.fluid_amb,
                         nn_conserve_momentum=nozzle_cons_m,
                         nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        generic_calc = hyram_overp.Generic_overpressure_method(jet_object, origin_at_orifice=True)
        for coord in generic_calc.origin:
            self.assertAlmostEqual(coord, 0)

    def test_bauwens_overp(self):
        p_rel = 60e5  # Pa
        release_fluid = Fluid(species={'h2': 1}, T=self.t_amb, P=p_rel)
        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', release_fluid)
        jet_object = Jet(release_fluid, self.orifice, self.fluid_amb,
                         nn_conserve_momentum=nozzle_cons_m,
                         nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        bauwens_calc = Bauwens_method(jet_object)
        locations = [(5, 0, 0)]
        overpressure = bauwens_calc.calc_overpressure(locations)[0]
        self.assertGreater(overpressure, 0)

    def test_bst_overp(self):
        p_rel = 60e5  # Pa
        release_fluid = Fluid(species={'h2': 1}, T=self.t_amb, P=p_rel)
        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', release_fluid)
        jet_object = Jet(release_fluid, self.orifice, self.fluid_amb,
                         nn_conserve_momentum=nozzle_cons_m,
                         nn_T=nozzle_t_param,
                         verbose=VERBOSE)

        mach_flame_speed = 5.2
        origin_at_orifice = True
        bst = BST_method(jet_object=jet_object, mach_flame_speed=mach_flame_speed, origin_at_orifice=origin_at_orifice)
        bst.flammable_mass = 1
        bst.energy = bst.calc_energy()
        self.assertEqual(bst.energy, 2*1*1.20E+08)

    def test_tnt_overp(self):
        p_rel = 60e5  # Pa
        release_fluid = Fluid(species={'h2': 1}, T=self.t_amb, P=p_rel)
        nozzle_cons_m, nozzle_t_param = misc_utils.convert_nozzle_model_to_params('yuce', release_fluid)
        jet_object = Jet(release_fluid, self.orifice, self.fluid_amb,
                         nn_conserve_momentum=nozzle_cons_m,
                         nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        tnt_factor = 0.03
        origin_at_orifice = True
        tnt = TNT_method(jet_object=jet_object, equivalence_factor=tnt_factor, origin_at_orifice=origin_at_orifice)

        tnt.flammable_mass = 1
        equiv_TNT_mass = tnt.calc_TNT_equiv_mass(tnt_factor, tnt.flammable_mass, tnt.heat_of_combustion)
        self.assertEqual(equiv_TNT_mass, 0.03*1*1.20E+08/4.68e6)


class GenericMethodTestCase(unittest.TestCase):
    """
    Tests of generic method functions for unconfined overpressure
    """
    def setUp(self):
        ambient_temperature = 300  # K
        jet_pressure = 60e5  # Pa
        orifice_diameter = 0.0254  # m
        dischage_coefficent = 1
        ambient_pressure = 101325  # Pa
        nozzle_model = 'yuce'
        release_fluid = Fluid(species='hydrogen',
                              T=ambient_temperature,
                              P=jet_pressure)
        orifice = Orifice(orifice_diameter, dischage_coefficent)
        ambient_fluid = Fluid(species='air',
                              P=ambient_pressure,
                              T=ambient_temperature)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, release_fluid)
        self.jet_object = Jet(release_fluid, orifice, ambient_fluid,
                              nn_conserve_momentum=nozzle_cons_momentum,
                              nn_T=nozzle_t_param,
                              verbose=VERBOSE)

    def test_origin_at_leak_orifice(self):
        generic_calc = hyram_overp.Generic_overpressure_method(self.jet_object,
                                                               origin_at_orifice=True)
        for coord in generic_calc.origin:
            self.assertAlmostEqual(coord, 0)

    def test_origin_not_at_leak_orifice(self):
        generic_calc = hyram_overp.Generic_overpressure_method(self.jet_object)
        self.assertGreater(generic_calc.origin[0], 0)
        self.assertGreater(generic_calc.origin[1], 0)
        self.assertEqual(generic_calc.origin[2], 0)

    def test_distance_no_locations(self):
        locations = []
        origin = (0, 0, 0)
        distances = hyram_overp.Generic_overpressure_method.calc_distance(locations,
                                                                          origin)
        self.assertEqual(len(distances), 0)

    def test_flammability_limits(self):
        lower_flammability_limit = 0.1
        upper_flammability_limit = 0.75
        flammability_limits = (lower_flammability_limit, upper_flammability_limit)
        generic_smaller_flam = hyram_overp.Generic_overpressure_method(self.jet_object,
                                                                       flammability_limits=flammability_limits)
        generic_larger_flam = hyram_overp.Generic_overpressure_method(self.jet_object)
        self.assertGreater(generic_larger_flam.flammable_mass, generic_smaller_flam.flammable_mass)


class BstMethodTestCase(unittest.TestCase):
    """
    Tests of different aspects of unconfined overpressure using BST method
    """
    def setUp(self):
        ambient_pressure = 101325  # Pa, ambient pressure
        ambient_temperature = 300  # K, jet exit temperature
        jet_pressure = 60e5  # Pa, jet exit pressure
        jet_diameter = 0.0254  #m, jet exit diameter
        ambient_fluid = Fluid(species='air', P=ambient_pressure, T=ambient_temperature)
        release_fluid = Fluid(species='hydrogen', T=ambient_temperature, P=jet_pressure)
        orifice_diameter = jet_diameter
        mach_flame_speed = 5.2
        origin_at_orifice = True
        nozzle_model = 'yuce'
        dischage_coefficent = 1

        orifice = Orifice(orifice_diameter, dischage_coefficent)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, release_fluid)
        jet_object = Jet(release_fluid, orifice, ambient_fluid,
                         nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param,
                         verbose=VERBOSE)

        self.BST_calc = BST_method(jet_object=jet_object, mach_flame_speed=mach_flame_speed, origin_at_orifice=origin_at_orifice)
        self.BST_calc.flammable_mass = 1
        self.BST_calc.energy = self.BST_calc.calc_energy()

    def test_calc_energy(self):
        self.assertEqual(self.BST_calc.energy, 2*1*1.20E+08)

    def test_calc_scaled_distance(self):
        distance = 1  #m
        scaled_distance = self.BST_calc.calc_scaled_distance(distance=distance)  # m/(J/Pa))^(1/3)
        self.assertEqual(scaled_distance, 1/(2*1*1.20E+08/101325)**(1/3))

    def test_get_scaled_overpressure(self):
        scaled_distance = 1  # m/(J/Pa)^(1/3)
        scaled_overpressure = self.BST_calc.get_scaled_overpressure(scaled_distance=scaled_distance)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_overpressure, 0.4, places=1)

    def test_calc_real_overpressure(self):
        scaled_overpressure = 10
        unscaled_overpressure = self.BST_calc.calc_unscaled_overpressure(scaled_overpressure=scaled_overpressure)
        self.assertEqual(unscaled_overpressure, scaled_overpressure*101325)

    def test_calc_overpressure(self):
        locations = [np.array([5, 0, 0])]  # m
        overpressure = self.BST_calc.calc_overpressure(locations=locations)[0]  # Pa
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(overpressure/202650, 202650/202650, places=0)

    def test_calc_overpressure_array(self):
        locations = [np.array([5, 0, 0]), np.array([0, 5, 0])]  # m
        overpressure = self.BST_calc.calc_overpressure(locations=locations)  # Pa
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        calculated_values = overpressure/202650
        test_values = np.array([202650/202650, 202650/202650])
        for calc_val, test_val in zip(calculated_values, test_values):
            self.assertAlmostEqual(calc_val, test_val, places=0)

    def test_get_scaled_impulse(self):
        scaled_distance = 1  # m/kg^(1/3)
        scaled_impulse = self.BST_calc.get_scaled_impulse(scaled_distance=scaled_distance)  # Pa*s*(m/s)/J^(1/3)/Pa^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_impulse, 0.03, places=0)

    def test_calc_real_impulse(self):
        scaled_impulse = 0.1
        unscaled_impulse = self.BST_calc.calc_unscaled_impulse(scaled_impulse=scaled_impulse)
        self.assertEqual(unscaled_impulse, scaled_impulse*(2*1*1.20E+08)**(1/3)*101325**(2/3) / 340)

    def test_calc_impulse(self):
        locations = [np.array([5, 0, 0])]  # m
        impulse = self.BST_calc.calc_impulse(locations=locations)[0]
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(impulse/283, 283/283, places=0)

    def test_changing_mach_flame_speed(self):
        mach_flame_speed = 0.7
        self.BST_calc.set_mach_flame_speed(mach_flame_speed=mach_flame_speed)
        self.assertEqual(self.BST_calc.mach_flame_speed, mach_flame_speed)
        mach_flame_speed = 1.4
        self.BST_calc.set_mach_flame_speed(mach_flame_speed=mach_flame_speed)
        self.assertEqual(self.BST_calc.mach_flame_speed, mach_flame_speed)

    def test_value_left_of_figure_data_returns_initial_value(self):
        for Mf in [0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2]:
            self.BST_calc.set_mach_flame_speed(mach_flame_speed=Mf)
            self.assertEqual(self.BST_calc.get_scaled_overpressure(scaled_distance=0.05),
                             self.BST_calc.scaled_peak_overpressure_data['scaled_overpressure_Mf' + str(Mf)][0])
            self.assertEqual(self.BST_calc.get_scaled_impulse(scaled_distance=0.005),
                             self.BST_calc.all_scaled_impulse_data['scaled_impulse_Mf' + str(Mf)][0])

    def test_value_right_of_figure_data_returns_final_value(self):
        for Mf in [0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2]:
            self.BST_calc.set_mach_flame_speed(mach_flame_speed=Mf)
            self.assertEqual(self.BST_calc.get_scaled_overpressure(scaled_distance=15),
                             self.BST_calc.scaled_peak_overpressure_data['scaled_overpressure_Mf' + str(Mf)][-1])
            self.assertEqual(self.BST_calc.get_scaled_impulse(scaled_distance=15),
                             self.BST_calc.all_scaled_impulse_data['scaled_impulse_Mf' + str(Mf)][-1])

    def test_calc_scaled_overpressure(self):
        overpressure = 13  # Pa
        scaled_overpressure = self.BST_calc.calc_scaled_overpressure(overpressure=overpressure)  # unitless
        self.assertEqual(scaled_overpressure, overpressure/101325)

    def test_get_scaled_distance_from_scaled_overpressure(self):
        scaled_overpressure = 10
        scaled_distance = self.BST_calc.get_scaled_distance_from_scaled_overpressure(scaled_overpressure=scaled_overpressure)  # m/(J/Pa)^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_distance, 0.2, places=1)

    def test_calc_unscaled_distance(self):
        scaled_distance = 1  # m/(J/Pa)^(1/3)
        unscaled_distance = self.BST_calc.calc_unscaled_distance(scaled_distance=scaled_distance)  # m
        self.assertEqual(unscaled_distance, 1*(2*1*1.20E+08/101325)**(1/3))

    def test_calc_distance_to_overpressure(self):
        overpressure = 202650  # Pa
        distance = self.BST_calc.calc_distance_to_overpressure(overpressure=overpressure)  # m
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(distance, 5, places=0)

    def test_calc_scaled_impulse(self):
        impulse = 13  # Pa*s
        scaled_impulse = self.BST_calc.calc_scaled_impulse(impulse=impulse)
        self.assertEqual(scaled_impulse, impulse*340/(2*1*1.20E+08)**(1/3)/101325**(2/3))

    def test_get_scaled_distance_from_scaled_impulse(self):
        scaled_impulse = 0.1
        scaled_distance = self.BST_calc.get_scaled_distance_from_scaled_impulse(scaled_impulse=scaled_impulse)  # m/(J/Pa)^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_distance, 0.27, places=0)

    def test_calc_distance_to_impulse(self):
        impulse = 284  # Pa*s
        distance = self.BST_calc.calc_distance_to_impulse(impulse=impulse)  # m
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(distance, 6, places=0)


class BauwensMethodTestCase(unittest.TestCase):
    """
    Tests of different aspects of unconfined overpressure using Bauwens method
    """
    def setUp(self):
        ambient_pressure = 101325  # Pa, ambient pressure
        ambient_temperature = 300  # K, jet exit temperature
        jet_pressure = 60e5  # Pa, jet exit pressure
        orifice_diameter = 0.0254  #m, jet exit diameter
        ambient_fluid = Fluid(species='air', P=ambient_pressure, T=ambient_temperature)
        release_fluid = Fluid(species='hydrogen', T=ambient_temperature, P=jet_pressure)
        nozzle_model='yuce'
        dischage_coefficent = 1

        orifice = Orifice(orifice_diameter, dischage_coefficent)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, release_fluid)
        jet_object = Jet(release_fluid, orifice, ambient_fluid,
                         nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param,
                         verbose=VERBOSE)

        self.Bauwens_calc = Bauwens_method(jet_object)

    def test_calc_overpressure(self):
        # test to check that result is non-zero
        locations = [(5, 0, 0)]
        overpressure = self.Bauwens_calc.calc_overpressure(locations)[0]
        self.assertGreater(overpressure, 0)

    def test_calc_impulse(self):
        # ensure that method returns a nan value due to not computing impulse
        locations = [(5, 0, 0)]
        impulse = self.Bauwens_calc.calc_impulse(locations)[0]
        self.assertTrue(np.isnan(impulse))

    def test_calc_energy(self):
        # test to check that result is non-zero
        self.assertGreater(self.Bauwens_calc.energy, 0)

    def test_calc_scaled_overpressure(self):
        scaled_distance = 3  # not 0, not 1
        scaled_overpressure = Bauwens_method.calc_scaled_overpressure_from_dist(scaled_distance)
        hand_calc_value = 0.0855
        self.assertAlmostEqual(scaled_overpressure, hand_calc_value, places=3)

    def test_calc_scaled_overpressure_zero_distance_error(self):
        self.assertRaises(ZeroDivisionError, Bauwens_method.calc_scaled_overpressure_from_dist, 0)

    def test_get_scaled_overpressure(self):
        scaled_distances = np.array([0, 1, 2])  # including zero and non-zero distances
        scaled_overpressures = self.Bauwens_calc.get_scaled_overpressure(scaled_distances)
        for scaled_overpressure in scaled_overpressures:
            self.assertGreater(scaled_overpressure, 0)


class TntMethodTestCase(unittest.TestCase):
    """
    Tests of different aspects of unconfined overpressure using TNT equivalence method
    """
    def setUp(self):
        self.ambient_pressure = 101325  # Pa, ambient pressure
        ambient_temperature = 300  # K, jet exit temperature
        jet_pressure = 60e5  # Pa, jet exit pressure
        jet_diameter = 0.0254  #m, jet exit diameter
        ambient_fluid = Fluid(species='air', P=self.ambient_pressure, T=ambient_temperature)
        release_fluid = Fluid(species='hydrogen', T=ambient_temperature, P=jet_pressure)
        orifice_diameter = jet_diameter
        equivalence_factor = 0.03
        origin_at_orifice = True
        nozzle_model='yuce'
        dischage_coefficent = 1

        orifice = Orifice(orifice_diameter, dischage_coefficent)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, release_fluid)
        jet_object = Jet(release_fluid, orifice, ambient_fluid,
                         nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param,
                         verbose=VERBOSE)

        self.TNT_method_calc = TNT_method(jet_object=jet_object,
                                          equivalence_factor=equivalence_factor,
                                          origin_at_orifice=origin_at_orifice)

        self.TNT_method_calc.flammable_mass = 1
        self.TNT_method_calc.equiv_TNT_mass = self.TNT_method_calc.calc_TNT_equiv_mass(equivalence_factor,
                                                                                       self.TNT_method_calc.flammable_mass,
                                                                                       self.TNT_method_calc.heat_of_combustion)

    def test_calc_TNT_mass(self):
        # Hand calculation based on 1 kg H2 with TNT blast energy 4.68 MJ/kg and yield of 3%
        equiv_TNT_mass = TNT_method.calc_TNT_equiv_mass(equivalence_factor=0.03,
                                                        flammable_mass=1,
                                                        heat_of_combustion=1.20E+08)
        self.assertEqual(equiv_TNT_mass, 0.03*1*1.20E+08/4.68e6)

    def test_calc_scaled_distance(self):
        distance = 1  # m
        scaled_distance = self.TNT_method_calc.calc_scaled_distance(distance=distance)  # m/kg^(1/3)
        self.assertEqual(scaled_distance, 1/((0.03*1*1.20E+08/4.68e6)**(1/3)))

    def test_get_scaled_overpressure(self):
        scaled_distance = 1  # m/kg^(1/3)
        scaled_overpressure = self.TNT_method_calc.get_scaled_overpressure(scaled_distance=scaled_distance)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_overpressure, 1.3e1, places=0)

    def test_calc_real_overpressure(self):
        scaled_overpressure = 13
        overpressure = self.TNT_method_calc.calc_unscaled_overpressure(scaled_overpressure=scaled_overpressure)  # Pa
        self.assertEqual(overpressure, scaled_overpressure*self.ambient_pressure)

    def test_calc_overpressure(self):
        locations = [np.array([1, 0, 0])]  # m
        overpressure = self.TNT_method_calc.calc_overpressure(locations=locations)[0]  # Pa
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(overpressure/1013250, 1013250/1013250, places=0)

    def test_get_scaled_impulse(self):
        scaled_distance = 3  # m/kg^(1/3)
        scaled_impulse = self.TNT_method_calc.get_scaled_impulse(scaled_distance=scaled_distance)  # Pa*s/kg^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_impulse, 9e-4, places=0)

    def test_calc_real_impulse(self):
        scaled_impulse = 13  # Pa*s/kg^(1/3)
        unscaled_impulse = self.TNT_method_calc.calc_unscaled_impulse(scaled_impulse=scaled_impulse)  # Pa*s
        self.assertEqual(unscaled_impulse, scaled_impulse*(0.03*1*1.20E+08/4.68e6)**(1/3))

    def test_calc_impulse(self):
        locations = [np.array([1, 0, 0])]  # m
        impulse = self.TNT_method_calc.calc_impulse(locations=locations)[0]
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(impulse/0.0021, 0.0021/0.0021, places=0)

    def test_value_left_of_figure_data_returns_initial_value(self):
        self.assertEqual(self.TNT_method_calc.get_scaled_overpressure(scaled_distance=0.05),
                         self.TNT_method_calc.scaled_peak_overP_data['scaled_overpressure'][0])
        self.assertEqual(self.TNT_method_calc.get_scaled_impulse(scaled_distance=0.05),
                         self.TNT_method_calc.scaled_impulse_data['scaled_impulse'][0])

    def test_value_right_of_figure_data_returns_final_value(self):
        self.assertEqual(self.TNT_method_calc.get_scaled_overpressure(scaled_distance=40),
                         self.TNT_method_calc.scaled_peak_overP_data['scaled_overpressure'][-1])
        self.assertEqual(self.TNT_method_calc.get_scaled_impulse(scaled_distance=40),
                         self.TNT_method_calc.scaled_impulse_data['scaled_impulse'][-1])

    def test_calc_scaled_overpressure(self):
        overpressure = 13
        scaled_overpressure = self.TNT_method_calc.calc_scaled_overpressure(overpressure=overpressure)  # unitless
        self.assertEqual(scaled_overpressure, overpressure/self.ambient_pressure)

    def test_get_scaled_distance_from_scaled_overpressure(self):
        scaled_overpressure = 100
        scaled_distance = self.TNT_method_calc.get_scaled_distance_from_scaled_overpressure(scaled_overpressure=scaled_overpressure)  # m/kg^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_distance, 0.3, places=1)

    def test_calc_unscaled_distance(self):
        scaled_distance = 1  # m/kg^(1/3)
        unscaled_distance = self.TNT_method_calc.calc_unscaled_distance(scaled_distance=scaled_distance)  # m
        self.assertEqual(unscaled_distance, 1*(0.03*1*1.20E+08/4.68e6)**(1/3))

    def test_calc_distance_to_overpressure(self):
        overpressure = 1013250  # Pa
        distance = self.TNT_method_calc.calc_distance_to_overpressure(overpressure=overpressure)  # m
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(distance, 1, places=0)

    def test_calc_scaled_impulse(self):
        impulse = 13  # Pa*s
        scaled_impulse = self.TNT_method_calc.calc_scaled_impulse(impulse=impulse)  # Pa*s/kg^(1/3)
        self.assertEqual(scaled_impulse, impulse/(0.03*1*1.20E+08/4.68e6)**(1/3))

    def test_get_scaled_distance_from_scaled_impulse(self):
        scaled_impulse = 1e-4  # Pa*s/kg^(1/3)
        scaled_distance = self.TNT_method_calc.get_scaled_distance_from_scaled_impulse(scaled_impulse=scaled_impulse)  # m/kg^(1/3)
        # Figure-estimation by hand
        self.assertAlmostEqual(scaled_distance, 31, places=0)

    def test_calc_distance_to_impulse(self):
        impulse = 0.0021  # Pa*s
        distance = self.TNT_method_calc.calc_distance_to_impulse(impulse=impulse)  # m
        # Hand calculation and figure estimation by hand
        # VERY rough approximate value
        self.assertAlmostEqual(distance, 0, places=0)


class TestJallaisOverpressureH2(unittest.TestCase):
    """
    Tests of hydrogen overpressure calculation from Jallais et al.
    """
    def setUp(self):
        ambient_pressure = 101325  # Pa
        ambient_temperature = 300  # K
        fluid_pressure = 60e5  # Pa
        orifice_diameter = 0.0254  # m
        ambient_fluid = Fluid(species='air', P=ambient_pressure, T=ambient_temperature)
        release_fluid = Fluid(species='hydrogen', T=ambient_temperature, P=fluid_pressure)
        nozzle_model='yuce'
        dischage_coefficent = 1
        orifice = Orifice(orifice_diameter, dischage_coefficent)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model)
        jet_object = Jet(release_fluid, orifice, ambient_fluid,
                         nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        self.jallais_calc = hyram_overp.JallaisOverpressureH2(jet_object)

    def test_reject_nonhydrogen_inputs(self):
        ambient_fluid = Fluid(species='air', P=101325, T=300)
        release_fluid = Fluid(species='methane', T=300, P=1e6)
        nozzle_model='yuce'
        orifice = Orifice(0.0254)
        nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model)
        jet_object = Jet(release_fluid, orifice, ambient_fluid,
                         nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param,
                         verbose=VERBOSE)
        self.assertRaises(ValueError, hyram_overp.JallaisOverpressureH2, jet_object)

    def test_calc_flame_speed(self):
        mass_flow_rate = 0.7  # kg/s
        flame_speed = hyram_overp.JallaisOverpressureH2.calc_flame_speed(mass_flow_rate)  # m/s
        self.assertEqual(flame_speed, 140)

    def test_calc_mach_flame_speed(self):
        flame_speed = 100  # m/s
        mach_flame_speed = self.jallais_calc.calc_mach_flame_speed(flame_speed)
        self.assertAlmostEqual(mach_flame_speed, 100/340)

    def test_get_mach_flame_speed_curve(self):
        mach_flame_speed = 0.29
        selected_mach_flame_speed = hyram_overp.JallaisOverpressureH2.get_mach_flame_speed_curve(mach_flame_speed)
        self.assertEqual(selected_mach_flame_speed, 0.35)

    def test_calc_overpressure(self):
        locations = [(5, 0, 0)]
        overpressures = self.jallais_calc.calc_overpressure(locations)
        for overpressure in overpressures:
            self.assertGreater(overpressure, 0)



if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import unittest
import json

from scipy.constants import pi

from hyram.phys import Fluid, Orifice, Flame
from hyram.utilities import misc_utils

import utils

"""
NOTE: if running from IDE like pycharm, make sure cwd is hyram/ and not hyram/tests.
"""

# Flags to enable command line output, pyplots, and text file output
VERBOSE = False
CREATE_PLOTS = False
CREATE_OUTPUT = False

# Absolute paths to input/output data
DATA_LOC = os.path.join(os.path.dirname(__file__), 'data')
OUTPUT_LOC = os.path.join(misc_utils.get_temp_folder(), 'validation-cryo')
OUTPUT_FILE = os.path.join(OUTPUT_LOC, 'Validation.txt')
LIMITS_FILE = os.path.join(DATA_LOC, 'Limits-cryo.json')


class Test_LabCryoData(unittest.TestCase):
    """
    Tests for flame length and heatflux, measured vs. calculated
    """
    def setUp(self):
        self.csv_data = utils.read_csv(os.path.join(DATA_LOC, 'cryo-data.csv'), header=0)

        with open(LIMITS_FILE) as f:
            self.limits = json.load(f)[__class__.__name__]

        air = Fluid(T=295, P=101325, species='air')

        self.test_ids = self.csv_data['TEST ID']
        data_length = len(self.test_ids)

        self.axial_dists = [.064, .22, .32, .42, .65]

        # Create lists for flamelength and heatflux
        self.flamelengths = []
        self.heatflux_qs = [[], [], [], [], []]
        for idx in range(data_length):
            diameter = self.csv_data['D (mm)'][idx] * 1e-3
            orifice = Orifice(diameter)

            temperature = self.csv_data['T (K)'][idx]
            pressure = self.csv_data['P (bar_g)'][idx] * 1e5 + 1e5
            gas = Fluid(T=temperature, P=pressure , species='h2')
            flame = Flame(gas, orifice, air, theta0=pi/2)

            self.flamelengths.append(flame.length())

            for idx, dist in enumerate(self.axial_dists):
                heatflux = flame.Qrad_multi(.2, dist, 0, RH=.8)
                self.heatflux_qs[idx].append(heatflux)

    def test_flamelength(self):
        title = "Lab Cryo Data - Flamelength"

        calc_flamelength = self.flamelengths
        exp_flamelength = self.csv_data['Fl']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='m',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

    def test_heatflux_dist1(self):
        idx = 0
        dist = self.axial_dists[idx]
        title = "Lab Cryo Data - Heatflux, Axial Distance = " + str(dist) + " m"

        calc_flamelength = self.heatflux_qs[idx]
        exp_flamelength = self.csv_data[f'Q{idx+1}']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='kW/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

    def test_heatflux_dist2(self):
        idx = 1
        dist = self.axial_dists[idx]
        title = "Lab Cryo Data - Heatflux, Axial Distance = " + str(dist) + " m"

        calc_flamelength = self.heatflux_qs[idx]
        exp_flamelength = self.csv_data[f'Q{idx+1}']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='kW/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

    def test_heatflux_dist3(self):
        idx = 2
        dist = self.axial_dists[idx]
        title = "Lab Cryo Data - Heatflux, Axial Distance = " + str(dist) + " m"

        calc_flamelength = self.heatflux_qs[idx]
        exp_flamelength = self.csv_data[f'Q{idx+1}']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='kW/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

    def test_heatflux_dist4(self):
        idx = 3
        dist = self.axial_dists[idx]
        title = "Lab Cryo Data - Heatflux, Axial Distance = " + str(dist) + " m"

        calc_flamelength = self.heatflux_qs[idx]
        exp_flamelength = self.csv_data[f'Q{idx+1}']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='kW/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")

    def test_heatflux_dist5(self):
        idx = 4
        dist = self.axial_dists[idx]
        title = "Lab Cryo Data - Heatflux, Axial Distance = " + str(dist) + " m"

        calc_flamelength = self.heatflux_qs[idx]
        exp_flamelength = self.csv_data[f'Q{idx+1}']

        error_limits = self.limits[title]
        error = utils.get_error(exp_vals=exp_flamelength,
                                calc_vals=calc_flamelength,
                                error_limits=error_limits,
                                units='kW/m^2',
                                msg=title,
                                output_filename=OUTPUT_FILE,
                                create_output=CREATE_OUTPUT,
                                verbose=VERBOSE)

        with self.subTest():
            self.assertLessEqual(error['Max Absolute Error'], error_limits['Max Absolute Error'], f"{title}: Maximum Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Absolute Error'], error_limits['Avg Absolute Error'], f"{title}: Average Absolute Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Max Percent Error'], error_limits['Max Percent Error'], f"{title}: Maximum Percent Error out of range")
        with self.subTest():
            self.assertLessEqual(error['Avg Percent Error'], error_limits['Avg Percent Error'], f"{title}: Average Percent Error out of range")
        with self.subTest():
            self.assertGreaterEqual(error['R2'], error_limits['R2'], f"{title}: R Squared Value out of range")


if __name__ == "__main__":
    if os.path.exists(OUTPUT_FILE):
        os.remove(OUTPUT_FILE)
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

"""
Probability distribution classes and helper funcs for convenience.
This file should NOT import from any QRA-specific code.
"""

import numpy as np
from scipy.stats import beta, uniform, norm


class DistributionWrapper(object):
    """
    Convenience class to build similar distributions (i.e. scipy continuous_ev).

    Parameters
    ----------
    a :
    b :
    distr_class : class
        scipy class of distribution, e.g. norm
    loc :
    scale :

    """

    def __init__(self, a, b, distr_class, loc=0, scale=1):
        self.name = None
        self.a = a
        self.b = b
        self.loc = loc
        self.scale = scale

        if distr_class == uniform:
            self.distribution = distr_class(a, b)
        else:
            self.distribution = distr_class(a, b, loc=self.loc, scale=self.scale)

        self.mean = float(self.distribution.mean())
        self.var = self.variance = float(self.distribution.var())
        self.rvs = self.distribution.rvs

    def __str__(self):
        return '{}, a={:.3g}, b={:.3g}'.format(self.name, self.a, self.b)


class BetaDistribution(DistributionWrapper):
    """ Beta distribution setup """

    def __init__(self, a, b, loc=0, scale=1):
        if a <= 0 or b <= 0:
            raise ValueError('Invalid beta distribution - parameters must be greater than 0')
        super().__init__(a, b, loc=loc, scale=scale, distr_class=beta)
        self.name = "Beta"


class UniformDistribution(DistributionWrapper):
    """ Uniform distribution setup """
    """

    """

    def __init__(self, a, b):
        if a >= b:
            raise ValueError('Uniform distribution parameter b must be >= a')
        super().__init__(a, b, distr_class=uniform)
        self.name = "Uniform"


class NormalDistribution(DistributionWrapper):
    """ Normal distribution setup """

    def __init__(self, a, b, loc=0, scale=1):
        if b <= 0:
            raise ValueError('Normal distribution parameter b must be > 0')
        super().__init__(a, b, loc=loc, scale=scale, distr_class=norm)
        self.name = "Normal"


class EVDistribution:
    """
    Expected Value distribution.

    Parameters
    ----------
    value :
    """

    def __init__(self, value, *args, **kwargs):
        self.name = "Expected Value"
        self.value = value
        self.distribution = ExpectedValue(value)

        self.mean = float(self.distribution.mean())
        self.var = self.variance = float(self.distribution.var())

    def __str__(self):
        return 'Expected Value = {:.3g}'.format(self.value)


class ExpectedValue:
    def __init__(self, value, *args, **kwargs):
        self.value = float(value)

    # def __getattribute__(self, attr):
    #     if attr == 'mean':
    #         return self.mean
    #     elif attr == 'ppf':
    #         return self.ppf
    #     elif attr == 'pdf':
    #         return self.pdf
    #     elif attr == 'cdf':
    #         return self.cdf
    #     elif attr == 'var':
    #         return self.var

    def mean(self):
        return self.value

    def ppf(self, percentile=None):
        return self.value

    def pdf(self, x):
        return x == self.value

    def cdf(self, x):
        return x >= self.value

    def var(self):
        return 0.


class LogNormDistribution(object):
    """

    """

    def __init__(self, mu, sigma):
        self.name = "Lognormal"
        self.mu = mu
        self.sigma = sigma

        # NOTE: does not use scipy lognorm since that can't accommodate s==shape==sigma==0
        self.distribution = None
        # NOTE: lognormal distribution will use geometric mean (median) rather than arithmetic mean
        self.mean = np.exp(self.mu)
        self.var = self.variance = (np.exp(self.sigma ** 2) - 1) * np.exp(2 * self.mu + self.sigma ** 2)
        self.rvs = None

    def __str__(self):
        return 'Lognormal, mu={:.3f}, sigma={:.3f}'.format(self.mu, self.sigma)


DISTRIBUTIONS = {
    'logn': LogNormDistribution,
    'beta': BetaDistribution,
    'expv': EVDistribution,
    'unif': UniformDistribution,
    'norm': NormalDistribution,
}


def get_distribution_class(distr_name):
    """ Retrieve correct distribution class based on string name """
    lname = str(distr_name).lower()
    if lname in ['lognorm', 'lognormal', 'logn']:
        return DISTRIBUTIONS['logn']

    elif lname in ['beta', 'bt']:
        return DISTRIBUTIONS['beta']

    elif lname in ['expv', 'ev', 'expectedvalue', 'expected_value', 'expected_val', 'exp_val']:
        return DISTRIBUTIONS['expv']

    elif lname in ['unif', 'uniform', 'uni']:
        return DISTRIBUTIONS['unif']

    elif lname in ['norm', 'normal', 'norml']:
        return DISTRIBUTIONS['norm']

    else:
        raise KeyError("Distribution {} not found. Available options are {}".format(distr_name, DISTRIBUTIONS.keys()))


def has_distribution(distr_name):
    """ Only checks whether ID matches a distribution; does not return it"""
    try:
        distr = get_distribution_class(distr_name)
        return True
    except Exception:
        return False


def get_distribution_options():
    return DISTRIBUTIONS.keys()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

import scipy.constants as spc

from hyram.qra import analysis as qra_analysis
from hyram.qra import component_failure
from hyram.qra.component_set import ComponentSet
from hyram.qra.data import component_data

VERBOSE = False


class QraBlendsTestCase(unittest.TestCase):
    """
    Test QRA with blends.
    """

    def setUp(self):
        rel_species = {"ch4": 0.96, 'n2': 0.04}
        rel_phase = "None"
        leak_sizes = None

        num_compressors = 0
        num_vessels = 0
        num_valves = 5
        num_instruments = 3
        num_joints = 35
        num_hoses = 1
        num_pipes = 20  # m
        num_filters = 0
        num_flanges = 0
        num_heat_exchangers = 0
        num_vaporizers = 0
        num_transfer_arms = 0
        num_extra1s = 0
        num_extra2s = 0

        component_sets = [
            ComponentSet('compressor', num_compressors, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['compressor'],
                         leak_sizes=leak_sizes),
            ComponentSet('vessel', num_vessels, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['vessel'],
                         leak_sizes=leak_sizes),
            ComponentSet('valve', num_valves, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['valve'],
                         leak_sizes=leak_sizes),
            ComponentSet('instrument', num_instruments, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['instrument'],
                         leak_sizes=leak_sizes),
            ComponentSet('joint', num_joints, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['joint'],
                         leak_sizes=leak_sizes),
            ComponentSet('hose', num_hoses, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['hose'],
                         leak_sizes=leak_sizes),
            ComponentSet('pipe', num_pipes, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['pipe'],
                         leak_sizes=leak_sizes),
            ComponentSet('filter', num_filters, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['filter'],
                         leak_sizes=leak_sizes),
            ComponentSet('flange', num_flanges, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['flange'],
                         leak_sizes=leak_sizes),
            ComponentSet('exchanger', num_heat_exchangers, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['exchanger'],
                         leak_sizes=leak_sizes),
            ComponentSet('vaporizer', num_vaporizers, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['vaporizer'],
                         leak_sizes=leak_sizes),
            ComponentSet('arm', num_transfer_arms, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['arm'],
                         leak_sizes=leak_sizes),
            ComponentSet('extra1', num_extra1s, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['extra1'],
                         leak_sizes=leak_sizes),
            ComponentSet('extra2', num_extra2s, species=rel_species, saturated_phase=rel_phase,
                         leak_frequency_lists=component_data.h2_gas_params['extra2'],
                         leak_sizes=leak_sizes)
        ]

        component_failure_set = component_failure.ComponentFailureSet(
                f_failure_override=None, num_vehicles=20, daily_fuelings=2, vehicle_days=250,
                noz_po_dist='beta', noz_po_a=0.5, noz_po_b=610415.5,
                noz_ftc_dist='expv', noz_ftc_a=0.002, noz_ftc_b=None,
                mvalve_ftc_dist='expv', mvalve_ftc_a=0.001, mvalve_ftc_b=None,
                svalve_ftc_dist='expv', svalve_ftc_a=0.002, svalve_ftc_b=None,
                svalve_ccf_dist='expv', svalve_ccf_a=0.000128, svalve_ccf_b=None,
                overp_dist='beta', overp_a=3.5, overp_b=310289.5,
                pvalve_fto_dist='logn', pvalve_fto_a=-11.74, pvalve_fto_b=0.67,
                driveoff_dist='beta', driveoff_a=31.5, driveoff_b=610384.5,
                coupling_ftc_dist='beta', coupling_ftc_a=0.5, coupling_ftc_b=5031)

        self.inputs = {
            "pipe_outer_diam": 0.375 * spc.inch,
            "pipe_thickness": 0.065 * spc.inch,
            "amb_temp": 15 + spc.zero_Celsius,
            "amb_pres": 101325,
            "rel_temp": 15 + spc.zero_Celsius,
            "rel_pres": 35 * spc.mega,
            "rel_phase": None,
            "facil_length": 20,  # m
            "facil_width": 12,  # m
            "immed_ign_probs": [0.008, 0.053, 0.23],
            "delayed_ign_probs": [0.004, 0.027, 0.12],
            "ign_thresholds": [0.125, 6.25],

            "occupant_input_list": [{
                'count': 9,
                'descrip': 'workers',
                'xdistr': 'uniform', 'xa': 1, 'xb': 20,
                'ydistr': 'dete', 'ya': 1, 'yb': None,
                'zdistr': 'unif', 'za': 1, 'zb': 12,
                'hours': 2000,
            }],
            "component_sets": component_sets,
            "component_failure_set": component_failure_set,
            "rel_species": rel_species,
            "leak_sizes": leak_sizes,

            "discharge_coeff": 1,
            "detection_credit": 0.9,
            "overp_method": 'bst',
            "tnt_factor": None,
            "bst_flame_speed": 0.35,
            "probit_thermal_id": 'eise', "exposure_time": 60, "probit_overp_id": 'head',
            "nozzle_model": 'yuce',
            "rel_angle": 0, "rel_humid": 0.89,
            "rand_seed": 3632850, "excl_radius": 0.01,
            "f_release_overrides": None,
            "verbose": VERBOSE, "output_dir": None, "create_plots": False,
        }

    # @unittest.skip
    def test_basic_ch4_n2(self):
        results = qra_analysis.conduct_analysis(**self.inputs)
        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 1.05e-5, places=2)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import random


class PositionGenerator:
    """
    Generates positions from a list of location distributions and parameters
    """
    def __init__(self, loc_distributions, exclusion_radius=0.01, seed=None):
        """
        Initializes position generator

        Parameters
        ----------
        loc_distributions : list
            List of location distributions
            Each location distribution within this list should be a list like:
                [n,
                 (xdist_type, xparam_a, xparam_b),
                 (ydist_type, yparam_a, yparam_b),
                 (zdist_type, zparam_a, zparam_b)]
            where
                n : int
                    Number of occupants for this distribution
                dist_type : str
                    One of 'deterministic', 'uniform', or 'normal'
                param_a and param_b depend on the distribution type:
                    For 'deterministic': param_a = value, param_b = None.
                    For 'uniform': param_a = min_val, param_b = maxval
                    For 'normal': param_a = mu, param_b = sigma
        exclusion_radius : float (optional)
            Minimum distance in meters from origin (0,0,0)
            in any direction that all generated positions must be;
            default value is 0.01 m
        seed : int or None (optional)
            Seed for random number generator;
            default value is None, will generate new seed

        Calculated
        ----------
        randgen : Random object
            Random number generator based on seed value
        totworkers : float
            Total number of workers for all distributions
        locs : list of locations
            List of locations of interest, one for each occupant,
            each location is a tuple of 3 coordinates (m):
            [(x1, y1, z1), (x2, y2, z2), ...]
        """
        self.seed = seed
        self.randgen = random.Random(seed)
        self.exclusion_radius = exclusion_radius
        self.loc_distributions = loc_distributions
        self.totworkers = self.calc_total_workers()
        self.gen_positions()

    def calc_total_workers(self):
        total_workers = 0
        for dist_group in self.loc_distributions:
            n = dist_group[0]
            if n < 0 or not isinstance(n, int):
                raise ValueError(f'Number of wokers ({n}) must be a non-negative integer')
            total_workers += n
        return total_workers

    def gen_positions(self):
        locations = []
        for dist_group in self.loc_distributions:
            n = dist_group[0]
            xyz_dists = dist_group[1:]
            for _ in range(n):
                (x, y, z) = self._gen_valid_position(xyz_dists)
                locations.append((x, y, z))
        self.locs = locations

    def _gen_valid_position(self, xyz_dists):
        max_guesses = 500
        for _ in range(max_guesses):
            (x, y, z) = self._gen_xyz_loc(xyz_dists)
            distance = (x**2 + y**2 + z**2)**(1/2)
            if distance > self.exclusion_radius:
                return x, y, z
        raise ValueError('Unable to produce desired number of valid positions outside exclusion radius')

    def _gen_xyz_loc(self, xyz_dists):
        """
        Generates an (x, y, z) location for one distribution-set

        Parameters
        ----------
        xyz_dists : list
            Distribution information for x, y, and z in that order

        Returns
        -------
        locs : list
            List of lists for x, y, and z locations
        """
        if len(xyz_dists) != 3:
            raise ValueError('List of distributions for a single location must contain 3 items')
        locs = []
        for dist in xyz_dists:
            locs.append(self._gen_loc(dist))
        return locs

    def _gen_loc(self, dist_params):
        """
        Generates a location on single dimension (x, y, or z) from distribution

        Parameters
        ----------
        dist_params : list or tuple
            Distribution information as
            (dist_type, param_a, param_b)
        """
        dist_type = dist_params[0]
        if dist_type in ['dete', 'det', 'deterministic']:
            return self._gen_deterministic(dist_params[1])
        elif dist_type in ['norm', 'normal']:
            return self._gen_normal(dist_params[1], dist_params[2])
        elif dist_type in ['unif', 'uni', 'uniform']:
            return self._gen_uniform(dist_params[1], dist_params[2])
        else:
            raise NotImplementedError(dist_type + ' distribution not implemented')
        
    def _gen_normal(self, mu, sigma):
        """
        Generates a number from normal distribution

        Parameters
        ----------
        mu : float
            Mean of normal distribution
        sigma : float
            Standard deviation of normal distribution

        Returns
        -------
        number : float
            Normally distributed number
        """
        return self.randgen.gauss(mu, sigma)

    def _gen_uniform(self, min_pos, max_pos):
        """
        Generates a number from uniform distribution

        Parameters
        ----------
        min_pos : float
            Minimum of range to generate from
        max_pos : float
            Maximum of range to generate from

        Returns
        -------
        number : float
            Uniformly distributed number
        """
        return self.randgen.uniform(min_pos, max_pos)
    
    def _gen_deterministic(self, val):
        """
        Generates a number of deterministic value
        
        Parameters
        ----------
        val : float
            Value to generate

        Returns
        -------
        number : float
            Deterministic number
        """
        return val

    def get_xlocs(self):
        """Get x locations"""
        return [location[0] for location in self.locs]

    def get_ylocs(self):
        """Get y locations"""
        return [location[1] for location in self.locs]

    def get_zlocs(self):
        """Get z locations"""
        return [location[2] for location in self.locs]
